#!/usr/bin/env python
# Tested on GITHub/Travis
# on your machine, just run pytest in this directory or execute it to get outputs
#

import pytest

from a2p2.jmmc import Catalog

oidbValidId = 38681

spicaCatName = "spica"
spicaCatName = "spica_2021_09_21"


class CustomCatalog(Catalog):
    def __init__(self, catalogName, username=None, password=None):
        Catalog.__init__(self, catalogName,username=username, password=password,
                         apiUrl="http://localhost:8080/exist/restxq/catalogs")
#                )


def test_same_id():
    has_same_id("oidb", oidbValidId)


def test_same_id2():
    has_same_id("oidb", oidbValidId)


def has_same_id(catname, id):
    c = CustomCatalog(catname)
    r = c.getRow(id)
    assert r["id"] == id
    print("We are ok with catalog %s id %s" % (catname, id))
    return

def test_protected():
    spica = CustomCatalog(spicaCatName, username="u", password="p")
    try:
        spica.getRow(1)
        spicaIsProtected = False
    except:
        spicaIsProtected = True

    assert spicaIsProtected


def test_public():
    oidb = CustomCatalog("oidb")
    try:
        oidb.getRow(oidbValidId)
        oidbIsPublic = True
    except:
        oidbIsPublic = False
    assert oidbIsPublic

def test_simple_update():
    c = CustomCatalog(spicaCatName)
    v=2
    c.updateRow(1, { "priority_pi" : v } )
    priority = c.getRow(1)["priority_pi"]
    assert priority==v

def test_simple_update_with_str_as_json_input():
    c = CustomCatalog(spicaCatName)
    v=2
    c.updateRow(1, '{"priority_pi":%d}'%v)
    priority = c.getRow(1)["priority_pi"]
    assert priority==v


def test_duplicated_col_update():
    c = CustomCatalog(spicaCatName)

    c.updateRow(1, {"priority_pi":1, "priority_pi":2})
    priority = c.getRow(1)["priority_pi"]
    assert priority==2

    # because sql certainly is case insensitive....
    try:
        c.updateRow(1, {"priority_pi":1, "PRIORITY_PI":2})
        assert False
    except:
        pass

def test_reject_bad_col():
    # TODO 
    c = CustomCatalog(spicaCatName)
    try:
        c.updateRow(1,{"wrong_key":42})
        assert False
    except:
        pass



#!/usr/bin/env python

__all__ = []

from a2p2.chara.gui import CharaUI
from a2p2.facility import Facility
import logging

HELPTEXT = "TODO update this HELP message in a2p2/chara/facility.py"

logger = logging.getLogger(__name__)

class CharaFacility(Facility):

    def __init__(self, a2p2client):
        Facility.__init__(self, a2p2client, "CHARA", HELPTEXT)
        self.charaUI = CharaUI(self)

    def processOB(self, ob):
        self.a2p2client.ui.addToLog(
            "OB received for '" + self.facilityName + "' interferometer")
        # show ob dict for debug
        self.a2p2client.ui.addToLog(str(ob), False)

        # performs operation
        self.consumeOB(ob)

        # give focus on last updated UI
        self.a2p2client.ui.showFacilityUI(self.charaUI)

    def consumeOB(self, ob):
        # for the prototype: just delegate handling to the GUI
        # we could imagine to store obs in a list and recompute a sorted
        # summary report e.g.
        self.charaUI.displayOB(ob)



#!/usr/bin/env python
# Tested on GITHub/Travis
# on your machine, just run pytest in this directory or execute it to get outputs
#

import os
import pytest

from a2p2.ob import OB
from a2p2 import A2p2Client


FIXTURE_DIR = os.path.join( os.path.dirname(os.path.realpath(__file__)) )

ALL_OBXML_VALUE_ERROR = pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR,"aspro-sample-bad-coords.obxml", ''),
    os.path.join(FIXTURE_DIR,"aspro-sample-bad-k.obxml", ''),
    os.path.join(FIXTURE_DIR,"aspro-sample.obxml", ''),
    os.path.join(FIXTURE_DIR,"aspro-sample-bad-insmode.obxml", ''),
)

ALL_OBXML_GENERAL_ERROR = pytest.mark.datafiles(
)

ALL_OBXML_NO_ERROR = pytest.mark.datafiles(
)


@ALL_OBXML_VALUE_ERROR
def test_obxml_value_error(datafiles):
    a2p2c=A2p2Client(True)

    for filepath in datafiles.listdir():
        file = str(filepath)
        errorLog = processOB(a2p2c, file)

        print("> ERROR log for %s"+ file)
        print(errorLog)
        print("< ERROR log \n\n")

        assert "Value error" in errorLog

def processOB(a2p2c, file):
    print("checking OB at %s location" % file)
    ob = OB(file)
    for obsc in ob.observationConfiguration:
        print("obsConfig.id=%s" % obsc.id)
    for obss in ob.observationSchedule.OB:
        print("obsSchedule.ref=%s" % str(obss.ref))
    #    print(ob)
    print("\n\n")
    a2p2c.clearErrors()
    a2p2c.processOB(ob)
    errorLog = "\n\r".join(a2p2c.getErrors())
    return errorLog



#!/usr/bin/env python

__all__ = []

import logging

from .catalogs import Catalog

logger = logging.getLogger(__name__)

#  TODO list and give access to whole services and their beta/alpha / pre-prod instances if existing

oidb = Catalog("oidb")



#!/usr/bin/env python

__all__ = []

import sys
import logging

from a2p2 import __release_notes__
from a2p2 import __version__

if sys.version_info[0] == 2:
    from Tkinter import *
    from tkMessageBox import *
    import ttk
else:
    from tkinter import *
    from tkinter.messagebox import *
    import tkinter.ttk as ttk

from distutils.version import LooseVersion

HELPTEXT = """This application provides the link between ASPRO (that you should have started) and interferometers facilities.

"""

logger = logging.getLogger(__name__)

class MainWindow():

    def __init__(self, a2p2client):

        self.a2p2client = a2p2client

        self.requestAbort = False

        self.window = Tk()

        try:
            dpi_value = self.window.winfo_fpixels('1i')
            # print("dpi: " + str(dpi_value));
            self.window.tk.call(
                'tk', 'scaling', '-displayof', '.', dpi_value / 72.0)
        except:
            print("Can not set tk scaling !")

        self.window.protocol("WM_DELETE_WINDOW", self._requestAbort)

        self.notebook = ttk.Notebook(self.window)

        self.logFrame = Frame(self.notebook)

        self.logtext = Text(self.logFrame)
        scroll = Scrollbar(self.logFrame, command=self.logtext.yview)
        self.logtext.configure(yscrollcommand=scroll.set)
        scroll.pack(side=RIGHT, fill=Y)
        self.logtext.pack(side=LEFT, fill=BOTH, expand=True)
        self.logFrame.pack(side=TOP, fill=BOTH, expand=True)

        self.helpFrame = Frame(self.notebook)
        self.helptabs = ttk.Notebook(self.helpFrame)
        self.helptabs.pack(side=TOP, fill=BOTH, expand=True)
        self.helpFrame.pack(fill=BOTH, expand=True)
        self.addHelp("A2P2", HELPTEXT)

        self.releaseNotesFrame = Frame(self.notebook)
        self.releaseNotesFrame.pack(fill=BOTH, expand=True)
        self.addReleaseNotes(self.releaseNotesFrame)

        # add tab and store index for later use in showFacilityUI
        self.tabIdx = {}
        self.registerTab("LOG", self.logFrame)
        self.registerTab("HELP", self.helpFrame)
        self.registerTab("RELEASE NOTES", self.releaseNotesFrame)
        self.notebook.select(self.tabIdx["LOG"])

        self.notebook.pack(side=TOP, fill=BOTH, expand=True)

        self.log_string = StringVar()
        self.log_string.set("log...")
        self.log = Label(self.window, textvariable=self.log_string, fg="blue")
        self.log.pack()

        self.status_bar = StatusBar(self.window)
        self.status_bar.pack(side=BOTTOM, fill=X)
        self.progress_value = self.status_bar.progress_value

    def __del__(self):
        self.window.destroy()

    def setSampId(self, id):
        if id:
            self.window.title("A2P2 v%s [%s]" % (__version__, id))
        else:
            self.window.title("A2P2 v" + __version__)

    def _requestAbort(self):
        self.requestAbort = True

    def addHelp(self, tabname, txt):
        frame = Frame(self.helptabs)
        widget = Text(frame, width=120)
        helpscroll = Scrollbar(frame, command=widget.yview)
        widget.configure(yscrollcommand=helpscroll.set)
        helpscroll.pack(side=RIGHT, fill=Y)
        widget.pack(side=LEFT, fill=BOTH, expand=True)
        frame.pack(side=TOP, fill=BOTH, expand=True)
        self.helptabs.add(frame, text=tabname)
        widget.insert(END, txt)

    def addReleaseNotes(self, frame):
        """
        Present in a ordered way the dictionnary version items.
        """
        txt = ""
        lvs = [LooseVersion(v) for v in __release_notes__]
        for lv in sorted(lvs, reverse=True):
            v=lv.vstring
            txt += "\nV " + v + " :"
            txt += "\n---------------"
            keys = []
            for k in sorted(__release_notes__[v]):
                if not ("TODO" in k):
                    keys.append(k)
            for k in sorted(__release_notes__[v]):
                if "TODO" in k:
                    keys.append(k)

            for e in keys:
                infos = __release_notes__[v][e]
                if infos:
                    txt += "\n  " + e + " : "
                    for i in infos:
                        txt += "\n    - " + i
            txt += "\n\n"

        widget = Text(frame, width=120)
        helpscroll = Scrollbar(frame, command=widget.yview)
        widget.configure(yscrollcommand=helpscroll.set)
        helpscroll.pack(side=RIGHT, fill=Y)
        widget.pack(side=LEFT, fill=BOTH, expand=True)
        widget.insert(END, txt)

    def registerTab(self, text, widget):
        self.notebook.add(widget, text=text)
        self.tabIdx[text] = len(self.tabIdx)

    def showFacilityUI(self, facilityUI):
        if not facilityUI.facility.facilityName in self.tabIdx.keys():
            self.registerTab(facilityUI.facility.facilityName, facilityUI)
        self.notebook.select(self.tabIdx[facilityUI.facility.facilityName])

    def quitAfterRunOnce(self):
        self.window.quit()

    def loop(self):
        self.window.after(50, self.quitAfterRunOnce)
        self.window.mainloop()
        self.update_status_bar()

    def innerloop(self):
        self.window.after(50, self.quitAfterRunOnce)
        self.window.mainloop()

    def update_status_bar(self):
        self.status_bar.set_label("SAMP", "SAMP: %s" %
                                  self.a2p2client.a2p2SampClient.get_status())
        self.status_bar.set_label(
            "API", "%s" % self.a2p2client.facilityManager.get_status())

    def get_api(self):
        return self.api

    def addToLog(self, text, displayString=True, level=logging.INFO):
        if displayString:
            self.log_string.set(str(text))
        self.logtext.insert(END, "\n" + str(text))
        self.logtext.see(END)
        self.showFrameToFront()
        self.refresh()
        logger.log(level,text)

    def refresh(self):
        self.window.update()

    def ShowErrorMessage(self, text):
        if self.a2p2client.fakeAPI:
            pass
        else:
            showerror("Error", text)
        self.addToLog("Error message")
        self.addToLog(text, False, logging.ERROR)

        # do also append to simple error stack (for testing purpose first)
        self.a2p2client.addError(text)

    def ShowWarningMessage(self, text):
        if self.a2p2client.fakeAPI:
            pass
        else:
            showwarning("Warning", text)
        self.addToLog("Warning message")
        self.addToLog(text, False, loggin.WARNING)

    def ShowInfoMessage(self, text):
        if self.a2p2client.fakeAPI:
            pass
        else:
            showinfo("Info", text)
        self.addToLog("Info message")
        self.addToLog(text, False, logging.INFO)

    def setProgress(self, perc):
        if perc > 1:
            perc = perc / 100.0
        self.progress_value.set(perc)
        if (perc <= 0) or (perc > 0.99):
            self.window.config(cursor="left_ptr")
        else:
            self.window.config(cursor="watch")
        self.innerloop()
        self.showFrameToFront()

    def showFrameToFront(self):
        self.window.attributes('-topmost', 1)
        self.window.attributes('-topmost', 0)


class StatusBar(Frame):

    def __init__(self, root, **kw):
        Frame.__init__(self, root, **kw)
        self.labels = {}

        self.progress_value = DoubleVar()
        self.progress_value.set(0.0)
        self.progressbar = ttk.Progressbar(
            self, orient='horizontal', length=200, maximum=1,
            variable=self.progress_value,
            mode='determinate')  # , from_=0, to=1, resolution=0.01,showvalue=0,takefocus=0)
        self.progressbar.pack(side=LEFT)

    def set_label(self, name, text='', side=RIGHT, width=0):
        if name not in self.labels:
            label = Label(self)
            label.pack(side=side, pady=0, padx=4, anchor=E)
            self.labels[name] = label
        else:
            label = self.labels[name]
        if width != 0:
            label.config(width=width)
        label.config(text=text)


class FacilityUI(Frame):

    def __init__(self, facility):
        Frame.__init__(self, facility.a2p2client.ui.notebook)
        # self.pack(fill=BOTH)
        self.facility = facility
        self.a2p2client = facility.a2p2client

    def addToLog(self, *args, **kwargs):
        """ Wrapper to log message in the common textfield """
        self.a2p2client.ui.addToLog(*args, **kwargs)

    def ShowErrorMessage(self, *args, **kwargs):
        self.a2p2client.ui.ShowErrorMessage(*args, **kwargs)

    def ShowWarningMessage(self, *args, **kwargs):
        self.a2p2client.ui.ShowWarningMessage(*args, **kwargs)

    def ShowInfoMessage(self, *args, **kwargs):
        self.a2p2client.ui.ShowInfoMessage(*args, **kwargs)

    def setProgress(self, perc):
        """ Wrapper to update progress bar """
        if perc > 1:
            perc = perc / 100.0
        if (perc <= 0) or (perc > 0.99):
            self.isIdle()
        else:
            self.isBusy()
        self.a2p2client.ui.setProgress(perc)

    def isBusy(self):
        """ Override me to lock specific widgets. """
        pass

    def isIdle(self):
        """ Override me to unlock specific widgets. """
        pass



#!/usr/bin/env python

__all__ = []

import logging
import json
import xml.etree.ElementTree as ET
from collections import defaultdict, namedtuple, OrderedDict

logger = logging.getLogger(__name__)

# https://stackoverflow.com/questions/2148119/how-to-convert-an-xml-string-to-a-dictionary-in-python
# see comment below for our custom mods on attributes naming
def etree_to_dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        # print("add d=%s"%str(d))
    if t.attrib:
        d[t.tag].update(('' + k, v)
                        for k, v in t.attrib.items())  # was '@' but can't be serialized by namedtuple
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d


class OB():
    """
    Give access to OBs sent by Aspro2 and provide some helper functions.
    You can use attributes to walk to associated configurations values, e.g.:
    - ob.interferometerConfiguration.name
    - ob.instrumentConfiguration
    - ob.observationConfiguration[]
    - ob.observationSchedule

    every values are string (must be converted for numeric values).

    """

    def __init__(self, url):
        # extract XML in elementTree

        e = ET.parse(url)
        d = etree_to_dict(e.getroot())
        # keep only content of subelement to avoid schema version change
        # '{http://www.jmmc.fr/aspro-ob/0.1}observingBlockDefinition'
        ds = d[list(d)[0]]  # -> version and name are lost
        # store attributes
        for e in ds.keys():
            # parse JSON into an object with attributes corresponding to dict keys.
            # We should probably avoid json use...
            o = json.loads(
                json.dumps(ds[e]), object_hook=lambda d: namedtuple(e, d.keys())(*d.values()))
            # observationConfiguration may be uniq but force it to be a list
            if "observationConfiguration" in e and not isinstance(o, list):
                setattr(self, e, [o])
            else:
                setattr(self, e, o)
        self.ds = ds

    def getFluxes(self, target):
        """
        Return a flux mesurements as dict (ordered dict by BVRIJHK ).
        """
        order = "BVRIJHK"
        fluxes = {}
        els = target._asdict()
        for k in els.keys():
            if k.startswith("FLUX_"):
                fluxes[k[5:]] = els[k]  # remove FLUX_ prefix for dict key

        return OrderedDict(sorted(fluxes.items(), key=lambda t: order.find(t[0])))

    def get(self, obj, fieldname, defaultvalue=None):
        if fieldname in obj._fields:
            return getattr(obj, fieldname)
        else:
            return defaultvalue

    def __str__(self):
        if True:
            import json
            return json.dumps(self.ds, indent=2)
        else:
            return str(self.ds)



#!/usr/bin/env python

__all__ = []

import logging

from .utils import JmmcAPI

logger = logging.getLogger(__name__)


PRODLABEL = {True: "production", False: "pre-prod"}


class Catalog():
    """ Get remote access to read and update catalogs exposed through JMMC's API.
    Credential can be explicitly given but .netrc file will be automagically used on 401.
    """

    # TODO move preprod toggle False by default
    def __init__(self, catalogName, username=None, password=None, preprod=True, apiUrl=None):
        self.catalogName = catalogName
        self.prod = not(preprod)
        # TODO manage here prod & preprod access points

        if apiUrl:
            self.apiUrl = apiUrl  # trust given url as catalogAPI if value is provided
        elif self.prod:
            self.apiUrl = ""  # no api in production yet
        else:
            self.apiUrl = "https://oidb-beta.jmmc.fr/restxq/catalogs"

        self.api = JmmcAPI(self.apiUrl, username, password)

        logger.info("Create catalog wrapper to access '%s' (%s API at %s)" %
                    (catalogName, PRODLABEL[self.prod], self.api.rootURL))

    def list(self):
        """ Get list of exposed catalogs on API associated to this catalog. """
        return self.api._get("")

    def metadata(self):
        """ Get catalog metadata """
        return self.api._get("/meta/%s" % self.catalogName)

    def pis(self):
        """ Get PIs from catalog and check for associated JMMC login """
        return self.api._get("/accounts/%s" % self.catalogName)

    def getRow(self, id):
        """ Get a single catalog record for the given id.
             usage: cat.getRow(42)
        """
        return self.api._get("/%s/%s" % (self.catalogName, id))

    def updateRow(self, id, values):
        """ Update record identified by given id and associated values.
            usage: cat.updateRows(42, {"col_a":"a", "col_b":"b" })
        """
        return self.api._put("/%s/%s" % (self.catalogName, id), values)

    def updateRows(self, values):
        """ Update multiple rows.
        Values must contain a list of dictionnary and each entry must contains id key among other columns.
            usage: updateRows([ { "id":42, "col_a":"a" }, { "id":24, "col_b":"b" } ])
        """

        # We may check befere sending payload that we always provide an id for every record
        return self.api._put("/%s" % (self.catalogName), values)

    def addRows(self, values):
        """ add multiple rows.
            Values is an array of row to add. id column values will be ignored.
            usage: addCatalogRows([ { "id":42, "col_a":"a" }, { "id":24, "col_b":"b" } ])
        """
        return self.api._post("/%s" % (self.catalogName), values)



#!/usr/bin/env python

__all__ = []

import logging

from .utils import JmmcAPI

logger = logging.getLogger(__name__)


#
# first release done for SAMP interoperability with Aspro2
#


def _model(model, output_mode=None):
    """ Returns a serialisation of given model to XML by default or another format given to output_mode value (only jsonlike at present time)."""
    if output_mode:
        return model
    else :
        return _xml_model(model)

def _xml_model(model):
    """ Rough conversion of dict to xml (Aspro2 namespaces)"""
    modeltype=model["type"]
    modelname=model["name"]

    params=""
    paramNames=[ k for k in model.keys() if not k in ("type", "name")]
    # at present time we must use Aspro2's approach with the same xml namespaces
    # see
    for p in paramNames:
        params+=f"""    <tm:parameter name="{modelname}_{p}" type="{p}"><value>{model[p]}</value></tm:parameter>\n"""

    return f"""<tm:model name="{modelname}" type="{modeltype}">\n{params}</tm:model>\n"""


class Models():
    """ Get analytical model's representations (in sync with Aspro2's ones).
    """
    SAMP_UCD_MODEL="meta.code.class;meta.modelled" # use it for colums that would be filled by models below

# generated code below from an Asprox file that contains a single star with list of all supported models
# cd a2p2/jmmc
# xml sel -o "from .models import _model" -n -n -b -t -m "//tm:model" -o "def " -v "@type" -o "( name, " -m "tm:parameter" -v "@type" -o "=" -v "value" -o ", " -b -o " output_mode=None):" -n -o '    """ ' -v "desc" -o ' """ ' -n -n -o '    return _model({' -m "tm:parameter" -o '"' -v "@type" -o '" : ' -v "@type" -o ", " -b -o '"name":name, "type":"' -v "@type" -o '"}, output_mode)' -n -n models_template.asprox > generated_models.py
# xml sel -t -o "    from .generated_models import " -m "//tm:model" -v "@type" -o ", "  -b -n models_template.asprox
# cd -
    from .generated_models import punct, disk, elong_disk, flatten_disk, circle, ring, elong_ring, flatten_ring, gaussian, elong_gaussian, flatten_gaussian, limb_quadratic



#!/usr/bin/env python

__all__ = []

import collections
import datetime
import json
import os
# import ast
import re
import logging

import numpy as np
from astropy.coordinates import SkyCoord

from a2p2.instrument import Instrument

logger = logging.getLogger(__name__)

class VltiInstrument(Instrument):

    def __init__(self, facility, insname):
        Instrument.__init__(self, facility, insname)
        self.facility = facility
        self.ui = facility.ui

        # use in latter lazy initialisation
        self.rangeTable = None
        self.ditTable = None

    def get(self, obj, fieldname, defaultvalue):
        if fieldname in obj._fields:
            return getattr(obj, fieldname)
        else:
            return defaultvalue


    def getSequence(self, ob):
    # We may look at ob.observationSchedule but at present tuime schedule still is computed on a2p2 side
        sci=[]
        cal=[]
        for observationConfiguration in ob.observationConfiguration:
            if 'SCIENCE' in observationConfiguration.type:
                sci.append(observationConfiguration)
            else:
                cal.append(observationConfiguration)
        if len(sci)+len(cal) <= 2: # return [CAL] [SCI]
            return cal+sci
        # else return CAL1 SCI CAL2 [SCI CAL3] [SCI CAL4]
        seq=[]
        for c in cal :
            seq.append(c)
            seq+=sci
        return seq[0:-1]


    # checkOB() must be defined by each instruments
    # def checkOB(self, ob, p2container=None):
    # consider dryMode if p2container is None else check again and submit on p2 side

    def submitOB(self, ob, p2container):
        api = self.facility.getAPI()
        ui = self.ui

        # create new container
        obsconflist = ob.observationConfiguration

        # Aspro2 always send SCI at first position (or CAL if no SCI)
        folderName = obsconflist[0].SCTarget.name
        folderName = re.sub('[^A-Za-z0-9]+', '_', folderName.strip())
        # prefix with username to make test folder clearer
        if p2container.isRoot() and self.facility.isTutorialAccount():
            folderName += "_" + self.facility.a2p2client.preferences.getP2UserCommentName()

        # do create a folder or concatenation only for SM or tutorial account
        # should we create a concatenation only if we have multiple objects
        if p2container.isRoot():
            # create a concatenation if service mode (even if we do not have any cal at this stage)
            if p2container.isServiceModeRun():
                # TODO fix bug when we keep the last selection that is a Concatenation. Concatenation can't be nested.
                # use parent container  or ask user to choose another container location ?
                folder, _ = api.createConcatenation(
                    p2container.containerId, folderName)
                ui.addToLog(f"concatenation '{folderName}' created")
            else:
                folder, _ = api.createFolder(p2container.containerId, folderName)
                ui.addToLog(f"folder '{folderName}' created")
            p2container.containerId = folder['containerId']
        else:
            ui.addToLog("concatenation or folder not created")

        ui.addToLog("OB checked / preparing  submission...")
        self.checkOB(ob, p2container)

    def getCoords(self, target, requirePrecision=True):
        """
        Format coordinates from given target to be VLTI compliant.
        Throws an exception if requirePrecision is true and given inputs have less than 3 (RA) or 2 (DEC) digits.
        """

        NAME = target.name

        RA = target.RA
        w = RA.rfind('.')
        l = len(RA)
        if l - w < 4 and requirePrecision:
            raise ValueError(
                "Object " + NAME + " has a too low precision in RA to be useable by VLTI, please correct with 3 or more digits.")
        if l - w > 4:
            RA = RA[0:w + 4]

        DEC = target.DEC
        w = DEC.rfind('.')
        l = len(DEC)
        if l - w < 3 and requirePrecision:
            raise ValueError(
                "Object " + NAME + " has a too low precision in DEC to be useable by VLTI,please correct with 2 or more digits.")
        if l - w > 4:
            DEC = DEC[0:w + 4]

        return RA, DEC

    def getPMCoords(self, target, defaultPMRA=0.0, defaultPMDEC=0.0):
        """
        Returns PMRA, PMDEC as float values rounded to 4 decimal digits. 0.0 is used as default if not present.
        """
        PMRA = self.get(target, "PMRA", defaultPMRA)
        PMDEC = self.get(target, "PMDEC", defaultPMDEC)
        return round(float(PMRA) / 1000.0, 4), round(float(PMDEC) / 1000.0, 4)

    def getFlux(self, target, flux):
        """
        Returns Flux as float values rounded to 3 decimal digits.

        flux in 'V', 'J', 'H'...
        """
        return round(float(getattr(target, "FLUX_" + flux)), 3)

    def getBaselineCode(self, baseline):
        # as of P104
        # take care if you change next branch order
        if "B2" in baseline:
            return "small"
        elif "G2" in baseline:
            return "medium"
        elif "J3" in baseline:
            return "large"
        elif "K0" in baseline:
            return "astrometric"
        elif "U" in baseline:
            return "UTs"
        else:
            raise ValueError(
                "Can't detect Interferometric Array type from given baseline : %s)" % (baseline))

    def checkIssVltiType(self, acqTSF):
        # TODO improve handling of this keyword using input from Aspro2's OB
        psection="p2.iss.vltitype"
        if self.facility.a2p2client.preferences.getConfigKeys(psection):
            vltitypePref=[v for v in self.facility.a2p2client.preferences.getConfigKeys(psection) if self.isInRange(acqTSF.tpl, "ISS.VLTITYPE", v) ]
            if vltitypePref:
                self.ui.addToLog(f"Using {psection} preference value : '{vltitypePref}'")
                acqTSF.ISS_VLTITYPE = vltitypePref


    def getA2p2Comments(self):
        return 'Generated by ' + self.facility.a2p2client.preferences.getP2UserCommentName() + \
               ' using ASPRO 2 (c) JMMC on ' + \
            datetime.datetime.now().isoformat()

    # ditTable and rangeTable are extracted from online doc:
    # -- https://www.eso.org/sci/facilities/paranal/instruments/gravity/doc/Gravity_TemplateManual.pdf
    # they are saved in json and displayed in the Help on application startup

    # k = 'GRAVITY_gen_acq.tsf'
    # using collections.OrderedDict to keep the order of keys:
    # rangeTable[k] = collections.OrderedDict({})
    # rangeTable[k]['SEQ.FI.HMAG']={'min':-10., 'max':20., 'default':0.0}
    # ...

    def getDitTable(self):
        if self.ditTable:
            return self.ditTable
        f = os.path.join(
            self.facility.getConfDir(), self.getName() + "_ditTable.json")
        self.ditTable = json.load(open(f))
        return self.ditTable

    def getDit(self, tel, spec, pol, K, dualFeed=False, showWarning=False):
        """
        finds DIT according to ditTable and K magnitude K

        'tel' in "AT" or "UT"
        'spec' in ditTable[tel].keys()
        'pol' in ditTable[tel][spec].keys()

        * a ValueError is thrown for out of range values *
        """
        # TODO add LOW mode  as new 'spec' entry for Gravity
        #        if showWarning and spec == "LOW":
        #            self.ui.ShowWarningMessage("DIT table does not provide LOW values. Using MED as workarround.")
        # if spec == "LOW":
        #    spec="HIGH"
        ditTable = self.getDitTable()
        mags = ditTable["AT"][spec][pol]['MAG']
        dits = ditTable["AT"][spec][pol]['DIT']
        if dualFeed:
            dK = ditTable["AT"]['Kdf']
        else:
            dK = 0.0
        if tel == "UT":
            dK += ditTable["AT"]['Kut']
        for i, d in enumerate(dits):
            if mags[i] < (K - dK) and (K - dK) <= mags[i + 1]:
                return d

        # handle out of bounds
        kmin = 1000
        kmax = -1000
        for i, d in enumerate(dits):
            kmin = min(kmin, mags[i] + dK)
            kmax = max(kmax, mags[i + 1] + dK)
        if kmin == K:
            return minDIT
        raise ValueError(
            "K mag (%f) is out of ranges [%f,%f]\n for this mode (tel=%s, spec=%s, pol=%s, dualFeed=%s)" % (
                K, kmin, kmax, tel, spec, pol, dualFeed))

    def getRangeTable(self):
        if self.rangeTable:
            return self.rangeTable
        f = os.path.join(
            self.facility.getConfDir(), self.getShortName() + "_rangeTable.json")
        # TODO use .tmp.json keys
        # using collections.OrderedDict to keep the order of keys:
        self.rangeTable = json.load(
            open(f), object_pairs_hook=collections.OrderedDict)
        return self.rangeTable

    def isInRange(self, tpl, key, value):
        """
        check if "value" is in range of keyword "key" for template "tpl"

        ValueError raised if key or tpl is not found.
        """
        rangeTable = self.getRangeTable()
        _tpl = ''
        # -- find relevant range dictionnary
        for k in rangeTable.keys():
            if tpl in [l.strip() for l in k.split(',')]:
                _tpl = k
        if _tpl == '':
            raise ValueError("unknown template '%s'" % tpl)
        if not key in rangeTable[_tpl].keys():
            raise ValueError(
                "unknown keyword '%s' in template '%s'" % (key, tpl))
        if 'min' in rangeTable[_tpl][key].keys() and \
                'max' in rangeTable[_tpl][key].keys():
            return value >= rangeTable[_tpl][key]['min'] and \
                value <= rangeTable[_tpl][key]['max']
        if 'list' in rangeTable[_tpl][key].keys():
            #            return value in rangeTable[_tpl][key]['list']
            if type(value) is list:
                for v in value:
                    if not v in rangeTable[_tpl][key]['list']:
                        return False
                    return True
            else:
                return value in rangeTable[_tpl][key]['list']
        if 'spaceseparatedlist' in rangeTable[_tpl][key].keys():
            ssl = rangeTable[_tpl][key]['spaceseparatedlist']
            for e in value.strip().split(" "):
                if not e in ssl:
                    return False
            return True
        # no range provided in tsf file
        return True

    def getRange(self, tpl, key):
        """
        returns range of keyword "key" for template "tpl"

        ValueError raised if key or tpl is not found.
        """
        rangeTable = self.getRangeTable()
        _tpl = ''
        # -- find relevant range dictionnary
        for k in rangeTable.keys():
            if tpl in [l.strip() for l in k.split(',')]:
                _tpl = k
        if _tpl == '':
            raise ValueError("unknown template '%s'" % tpl)
        if not key in rangeTable[_tpl].keys():
            raise ValueError(
                "unknown keyword '%s' in template '%s'" % (key, tpl))
        if 'min' in rangeTable[_tpl][key].keys() and \
                'max' in rangeTable[_tpl][key].keys():
            return (rangeTable[_tpl][key]['min'], rangeTable[_tpl][key]['max'])
        if 'list' in rangeTable[_tpl][key].keys():
            return rangeTable[_tpl][key]['list']
        if 'spaceseparatedlist' in rangeTable[_tpl][key].keys():
            for e in value.split(" "):
                return rangeTable[_tpl][key]['spaceseparatedlist']

    def getDefault(self, tpl, key):
        """
        returns the default value if any exists for given "key" in template "tpl", else None

        ValueError raised if key or tpl is not found .
        """
        rangeTable = self.getRangeTable()
        _tpl = ''
        # -- find relevant range dictionnary
        for k in rangeTable.keys():
            if tpl in [l.strip() for l in k.split(',')]:
                _tpl = k
        if _tpl == '':
            raise ValueError("unknown template '%s'" % tpl)
        if not key in rangeTable[_tpl].keys():
            raise ValueError(
                "unknown keyword '%s' in template '%s'" % (key, tpl))
        if 'default' in rangeTable[_tpl][key].keys() :
            return rangeTable[_tpl][key]['default']
        return None

    def getRangeDefaults(self, tpl, ):
        """
        returns a dict of keywords/default values for template "tpl"

        ValueError raised if tpl is not found.
        """
        rangeTable = self.getRangeTable()
        _tpl = ''
        # -- find relevant range dictionnary
        for k in rangeTable.keys():
            if tpl in [l.strip() for l in k.split(',')]:
                _tpl = k
        if _tpl == '':
            raise ValueError("unknown template '%s'" % tpl)
        res = {}
        for key in rangeTable[_tpl].keys():
            if 'default' in rangeTable[_tpl][key].keys():
                res[key] = rangeTable[_tpl][key]["default"]
        return res


    def getSkyDiff(self, ra, dec, ftra, ftdec):
        science = SkyCoord(ra, dec, frame='icrs', unit='deg')
        ft = SkyCoord(ftra, ftdec, frame='icrs', unit='deg')
        ra_offset = (science.ra - ft.ra) * np.cos(ft.dec.to('radian'))
        dec_offset = (science.dec - ft.dec)
        return [ra_offset.deg * 3600 * 1000, dec_offset.deg * 3600 * 1000]  # in mas

    def getSiderealTimeConstraints(self, LSTINTERVAL):
        # by default, above 40 degree. Will generate a WAIVERABLE ERROR if not.
        # we could warn before ?
        intervals = []
        if LSTINTERVAL:
            for lst in LSTINTERVAL:
                lsts = lst.split('/')
                # p2 seems happy with endlst < startlst
                # a = SkyCoord(lstStartSex+' +0:0:0',unit=(u.hourangle,u.deg))
                # b = SkyCoord(lstEndSex+' +0:0:0',unit=(u.hourangle,u.deg))
                # if b.ra.deg < a.ra.deg:
                # api.saveSiderealTimeConstraints(obId,[ {'from': lstStartSex, 'to': '00:00'},{'from': '00:00','to': lstEndSex}], stcVersion)
                # else:
                lstStartSex = lsts[0]
                lstEndSex = lsts[1]
                intervals.append({'from': lstStartSex, 'to': lstEndSex})
        return intervals

    def saveSiderealTimeConstraints(self, api, obId, LSTINTERVAL):
        # wait for next Aspro release to only send constraints set by user
        return
        intervals = self.getSiderealTimeConstraints(LSTINTERVAL)
        if intervals:
            sidTCs, stcVersion = api.getSiderealTimeConstraints(obId)
            api.saveSiderealTimeConstraints(obId, intervals, stcVersion)

    def getHelp(self):
        s = self.getName()
        s_name = s
        if s_name == "GRAVITY":
            from a2p2.vlti.gravity import Gravity
            s += "\n\n GravityRangeTable: \n"
            s += Gravity.formatRangeTable(self)
            s += "\n\nGravityDitTable:\n"
            s += Gravity.formatDitTable(self)
        if s_name == "PIONIER":
            from a2p2.vlti.pionier import Pionier
            s += "\n\n PionierRangeTable: \n"
            s += Pionier.formatRangeTable(self)
            s += "\n\nPionierDitTable:\n"
            s += Pionier.formatDitTable(self)
        if s_name == "MATISSE_LM":  # "MATISSE_LM" "MATISSE_N"
            from a2p2.vlti.matisse import Matisse
            s += "\n\n MatisseRangeTable: \n"
            s += Matisse.formatRangeTable(self)
            s += "\n\nMatisseDitTable:\n"
            s += Matisse.formatDitTable(self)

        return s

    def showP2Response(self, response, ob, obId):
        if response['observable']:
            msg = 'OB ' + \
                  str(obId) + ' submitted successfully on P2\n' + \
                  ob['name'] + ' is OK.'
        else:
            msg = 'OB ' + str(obId) + ' submitted successfully on P2\n' + ob[
                'name'] + ' has WARNING.\n see LOG for details.'
        self.ui.addToLog('\n')
        #self.ui.ShowInfoMessage(msg)
        self.ui.addToLog(msg)
        self.ui.addToLog('\n'.join(response['messages']) + '\n\n')


# TemplateSignatureFile
# use new style class to get __getattr__ advantage


class TSF(object):

    def __init__(self, instrument, tpl):
        self.tpl = tpl
        self.instrument = instrument
        supportedTpl = instrument.getRangeTable().keys()

        # DO not init with default values for every keywords so we only transfert changed values
        # this also reduce the risk to transmit some  incompatible keywords accross various IP versions
        # a TODO will come to test accross dicts get from P2 and not in our json config
        # self.tsfParams = self.instrument.getRangeDefaults(tpl)
        self.tsfParams = {}

        self.__initialised = True
        # after initialisation, setting attributes is the same as setting an item

    def set(self, key, value, checkRange=True):
        if checkRange:
            if not self.instrument.isInRange(self.tpl, key, value):
                raise ValueError(
                    "Parameter value (%s) is out of range for keyword %s in template %s " % (str(value), key, self.tpl))
        # TODO check that key is valid when checkRange is False
        self.tsfParams[key] = value

    def get(self, key):
        # TODO offer to get default value
        return self.tsfParams[key]

    def getDict(self):
        return self.tsfParams

    def getP2Name(self):
        return self.tpl[0:-4]

    def __getattr__(self, name):  # called for non instance attributes (i.e. keywords)
        rname = name.replace('_', '.')
        if rname in self.tsfParams:
            return self.tsfParams[rname]
        else:
            raise AttributeError(
                "unknown keyword '%s' in template '%s'" % (rname, self.tpl))

    def __setattr__(self, name, value):
        if name in self.__dict__:  # any normal attributes are handled normal
            return object.__setattr__(self, name, value)
        if not '_TSF__initialised' in self.__dict__:  # this test allows attributes to be set in the __init__ method
            return object.__setattr__(self, name, value)

        # continue with keyword try
        rname = name.replace('_', '.')
        self.set(rname, value)

    def __str__(self):
        buffer = "TSF values (%s) : \n" % (self.tpl)
        for e in self.tsfParams:
            buffer += "    %30s : %s\n" % (e, str(self.tsfParams[e]))
        return buffer


class FixedDict(object):

    def __init__(self, keys):
        self.myKeys = keys
        self.myValues = {}

        # Note: we could enhance code for checking + default value support such as TSF
        # Note: keys may be synchronized with p2

        self.__initialised = True
        # after initialisation, setting attributes is the same as setting an
        # item

    def getDict(self):
        return self.myValues

    def __getattr__(self, name):  # called for non instance attributes (i.e. keywords)
        rname = name.replace('_', '.')
        if rname in self.myValues:
            return self.myValues[rname]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in self.__dict__:  # any normal attributes are handled normally
            # print("set1 %s to %s"%(name, str(value)))
            return object.__setattr__(self, name, value)
        # this test allows attributes to be set in the __init__ method
        if not '_FixedDict__initialised' in self.__dict__:
            # print("set2 %s to %s"%(name, str(value)))
            return object.__setattr__(self, name, value)

        # continue with keyword try
        # print("set3 %s to %s"%(name, str(value)))
        rname = name.replace('_', '.')
        if rname in self.myKeys:
            self.myValues[name] = value
        else:
            raise ValueError(
                "keyword %s is not part of supported ones %s " % (name, self.myKeys))

    def __str__(self):
        buffer = "%s values:\n" % (type(self).__name__)
        for e in self.myValues:
            buffer += "    %30s : %s\n" % (e, str(self.myValues[e]))
        return buffer


class OBTarget(FixedDict):

    def __init__(self):
        FixedDict.__init__(
            self, ('name', 'ra', 'dec', 'properMotionRa', 'properMotionDec'))


class OBConstraints(TSF):

    def __init__(self, instrument):
        TSF.__init__(self, instrument, "instrumentConstraints.tsf")
        # WORKARROUND missing param in p2 json but valid and used in our code
        instrument.getRangeTable()[self.tpl]["name"]={}
